#![allow(dead_code)]

mod engine;

use component_macro::Component;
use engine::Component;
use engine::QuwiQuwi;
use engine::System;
use std::rc::Rc;

#[derive(Component)]
struct Sprite {
    sprite: engine::Sprite,
}

#[derive(Component)]
struct Position(f64, f64);

#[derive(Clone, Component)]
struct Velocity(f64, f64);

struct VelocitySystem;
impl System for VelocitySystem {
    fn on_update(&self, ctx: &mut engine::Context, delta: f64) -> Result<(), engine::Error> {
        for id in engine::Quwi::<(Velocity, Position)>::new().run(&ctx) {
            let vel = ctx.entity_component::<Velocity>(id).clone();
            let Position(x, y) = ctx.entity_component::<Position>(id);
            *x += vel.0 * delta;
            *y += vel.1 * delta;
        }
        Ok(())
    }
}

#[derive(Component)]
struct Gravity;

struct GravitySystem;
impl System for GravitySystem {
    fn on_update(&self, ctx: &mut engine::Context, delta: f64) -> Result<(), engine::Error> {
        for id in engine::Quwi::<(Gravity, Velocity)>::new().run(&ctx) {
            let Velocity(_, y) = ctx.entity_component::<Velocity>(id);
            *y = if *y < 200.0 { *y + 100.0 * delta } else { *y };
        }
        Ok(())
    }
}

struct SpriteRenderer;
impl System for SpriteRenderer {
    fn on_update(&self, ctx: &mut engine::Context, _delta: f64) -> Result<(), engine::Error> {
        for id in engine::Quwi::<(Sprite, Position)>::new().run(&ctx) {
            let &mut Position(x, y) = ctx.entity_component::<Position>(id);
            let sprite = ctx.entity_component::<Sprite>(id).sprite;

            ctx.draw_sprite(sprite, x as i32, y as i32)?;
        }
        Ok(())
    }
}

#[derive(Component)]
struct PlayerMovement;

struct PlayerMovementSystem;
impl System for PlayerMovementSystem {
    fn on_update(&self, ctx: &mut engine::Context, _delta: f64) -> Result<(), engine::Error> {
        for id in engine::Quwi::<(PlayerMovement, Velocity)>::new().run(&ctx) {
            let d_down = ctx.key_pressed(engine::Keycode::D);
            let a_down = ctx.key_pressed(engine::Keycode::A);
            let Velocity(x, _) = ctx.entity_component::<Velocity>(id);
            *x = if d_down && !a_down {
                100.0
            } else if !d_down && a_down {
                -100.0
            } else {
                0.0
            };
        }
        Ok(())
    }
}

fn main() {
    let mut game = engine::Game::new().unwrap();

    let mut context = game.context();
    context.add_system(Rc::new(VelocitySystem));
    context.add_system(Rc::new(SpriteRenderer));
    context.add_system(Rc::new(GravitySystem));
    context.add_system(Rc::new(PlayerMovementSystem));
    let sprite = context.load_sprite("textures/player.png").unwrap();
    context.spawn(vec![
        Box::new(Sprite { sprite }),
        Box::new(Position(16.0, 140.0)),
        Box::new(Velocity(0.0, -150.0)),
        Box::new(Gravity),
        Box::new(PlayerMovement),
    ]);
    game.run();
}
